# SeqGeq / Seurat analysis

# Script modified from Satija lab's work for the R Seurat package.
# Script modified by Miguel Velazquez, Ian Taylor, and Josh Luthy at FlowJo
# For use w/ SeqGeq 
# Seurat is developed and maintained by the Satija lab. 
# Code is  released in GPL (3.0) license.

#########################
## TRY TO AUTO INSTALL ##
#########################

# Install function for packages
loadPackages <-function(...) {
  libs<-unlist(list(...))
  req<-unlist(lapply(libs,require,character.only=TRUE))
  need<-libs[req==FALSE]
  if(length(need)>0){
    install.packages(need, repos="https://cran.r-project.org")
    lapply(need,require,character.only=TRUE)
  }
}

loadBioc<-function(...) {
  libs<-unlist(list(...))
  req<-unlist(lapply(libs,require,character.only=TRUE))
  need<-libs[req==FALSE]
  if(length(need)>0){
    BiocManager::install(need)
    lapply(need,require,character.only=TRUE)
  }
}

loadPackages(c('gplots','RColorBrewer','dplyr','data.table','Matrix','Seurat','cowplot','reshape2','ggplot2',
                'tidyr','sctransform','tools','BiocManager','viridisLite','future'))

loadBioc('multtest')

# Let's try for a speed boost if your machine has the power
numCores <- as.integer(future::availableCores())
if (numCores <= 2 ){
  plan("sequential")
  print("no extra cores, leave it at single core")
}else if ((numCores >= 3) && (numCores <= 7)){
  plan("multisession", workers = 2)
  print("set at two cores")
}else if ((numCores >= 8) && (numCores <= 16 )){
  plan("multisession", workers = 3)
  print("set at three cores")
} else if (numCores >= 17){
  plan("multisession", workers = 4)
  print("set at four cores")
}

# Let's just use two R instances
# plan("multisession", workers = 2)
# Setting max size to 4GB for each instance
options(future.globals.maxSize = 4000*1024^2)

sessionInfo()
## Other Functions ##
fig2 <- function(DF) {
  ggplot(DF, aes(y = Features, x = Type, fill = Fill)) +
    theme(axis.text.x = element_text(angle = 45, hjust = 1, vjust = 0.9, size = 10),
          axis.text.y = element_text(hjust = 0.9, vjust = 0.5, size = 10),
          legend.title = element_text(size = 8, face="bold"),
          legend.key.size = unit(2,"line"),
          legend.text = element_text(size = 8)) +
    geom_tile() +
    theme(plot.margin = margin(1, 1, 1, 3, "cm")) +
    labs(x = "", y = "") +
    scale_fill_viridis_c(name = "Rank Cell Call")
}

######################
######################
getSeuratVersion <- function(){
  pkgVersion <- getNamespaceVersion("Seurat")
  pkgVersion <- file_path_sans_ext(pkgVersion)
  pkgVersion <- as.numeric(pkgVersion)
}
yourVersion <- getSeuratVersion()
paste0("The installed version of Seurat is: ", yourVersion)

## For Error Messaging
stopAndReturnError <- function(message, ...) {
  cat(message, file=paste0(outfileDirectory, "/err.txt"))
  stop(message, ...)
}

ReturnErrorMessage <- function(message, ...) {
  cat(message, file=paste0(outfileDirectory, "Plotting_Error_Message_from_R.txt"))
}

#%%%%%%%%%%%%%%%%%%%%%%%#
## Read Data from Java ##
#%%%%%%%%%%%%%%%%%%%%%%%#
adtParNames <- c(ADT_PARAMS_LIST)
rezza <- SG_RESOLUTION
runID <- "SG_SEURAT_RUN_ID"
model <- "SG_MODEL"
runQC <- SG_RUN_QC_BOOL
minCellCutoffBool <- SG_CELL_QC_BOOL
minGeneCutoffBool <- SG_GENE_QC_BOOL
filterCells <- SG_FILTER_CELLS_BOOL
scTransform <- SG_SCTRANSFORM
logNormalizeBool <- SG_LOG_NORMALIZE_BOOL
runTSNE <- SG_RUN_TSNE
runUMAP <- SG_RUN_UMAP
classifyClusters <- RUN_CLASSIFIER
minGenePercent <- SG_MIN_GENE_PERCENT
cellPercent <- SG_CELL_QC_PERCENT
runBatchCorrection <- SG_BATCH_CORRECT
runWeightedNearestNeighbors <- SG_MULTIMODAL
outfileDirectory <- "SG_OUTPUT_DIRECTORY_PATH"
setwd(outfileDirectory)

# Read CSV from SeqGeq
importedCSVfile <- fread(file= "SG_SAMPLE_DATA_PATH",
                     sep = ",", header = TRUE, check.names = FALSE)
featureAnnotationsFrame <- data.frame(importedCSVfile$Gene)
featureAnnos <- as.character(importedCSVfile$Gene)
cellAnnotationsFrame <- data.frame(colnames(importedCSVfile))
if(any(names(importedCSVfile) == "Gene")) {
  cellAnnos <- as.character(colnames(importedCSVfile))
  cellAnnos <- cellAnnos[cellAnnos != "Gene"]

  cellAnnotationsFrame <- data.frame(colnames(importedCSVfile))
  rowsToKeep <- importedCSVfile[, which(names(importedCSVfile) != "Gene")]
  cellAnnotationsFrame <- data.frame(cellAnnotationsFrame[rowsToKeep,])
  colnames(cellAnnotationsFrame) <- paste("CellId")
}
rm(rowsToKeep)
importedCSVfile <- as.data.frame(importedCSVfile)
rownames(importedCSVfile) <- featureAnnos
importedCSVfile <- dplyr::select(importedCSVfile, -Gene)
# Some data has numbers as cellIDs only, append a string to cellIds, remove on export.
colnames(importedCSVfile) <- paste("AppID", colnames(importedCSVfile), sep="_")

if(runWeightedNearestNeighbors) {
  tCSVfile <- t(importedCSVfile)
  tCSVfile <- as.data.frame(tCSVfile)
  adtSubset <- select(tCSVfile, adtParNames)
  transcriptSubset <- select(tCSVfile, !adtParNames)
  adtSubset <- t(adtSubset)
  transcriptSubset <- t(transcriptSubset)

}else {
  tCSVfile <- importedCSVfile
}
#%%%%%%%%%%%%%%%%%%%%%%%#
## Run QC ##
#%%%%%%%%%%%%%%%%%%%%%%%#
if (!runBatchCorrection){ # if not performing batch correction, proceed with OG workflow.
  if (runQC) {
    # Boolean for include genes w/ min detected expression
    if (minCellCutoffBool) {
      minCellPercentVal <- cellPercent
      minCellCutoff <- nrow(importedCSVfile) * minCellPercentVal
    } else { minCellCutoff = cellPercent }

    if(minGeneCutoffBool) {
      minGeneCutoff <- ncol(importedCSVfile) * minGenePercent
    } else { minGeneCutoff <- minGenePercent }

    if (logNormalizeBool) {
      logNormalizeData <- "logNormalize"
    }
    if (runWeightedNearestNeighbors) {
      # create Seurat objects for protein and rna measurements
      seuratSeqGeq <- CreateSeuratObject(transcriptSubset,
                                         assay = 'RNA',
                                         min.cells = minCellCutoff,
                                         min.features = minGeneCutoff)
      # Add Protein measurements as a new assay 'ADT'
      seuratSeqGeq[["ADT"]] <- CreateAssayObject(adtSubset[, colnames(x = seuratSeqGeq)])
    } else {
      seuratSeqGeq <- CreateSeuratObject(tCSVfile,
                                         min.cells = minCellCutoff,
                                         min.features = minGeneCutoff)
    }
    rm(importedCSVfile)
  }else {
    # No QC, just create Seurat object
    if(runWeightedNearestNeighbors) {
      # create Seurat objects for protein and rna measurements
      seuratSeqGeq <- CreateSeuratObject(transcriptSubset,
                                         assay = 'RNA')
      # Add Protein measurements as a new assay 'ADT'
      seuratSeqGeq[["ADT"]] <- CreateAssayObject(counts = adtSubset)
    } else {
      # No QC, just create Seurat object
      seuratSeqGeq <- CreateSeuratObject(tCSVfile)
    }
    rm(importedCSVfile)
  }

  checkMitoGenes = TRUE
  if(yourVersion < 5.0){
    if (checkMitoGenes){
      hu.mito.genes <- grep(pattern = "^MT-", x = rownames(x = seuratSeqGeq@assays[["RNA"]]@data), value = TRUE)
      if (length(hu.mito.genes) < 1)
      { # Possibly mouse
        mo.mito.genes <- grep(pattern = "^mt-", x = rownames(x = seuratSeqGeq@assays[["RNA"]]@data), value = TRUE)
        DefaultAssay(seuratSeqGeq) <- 'RNA'
        seuratSeqGeq <- PercentageFeatureSet(seuratSeqGeq, pattern = "^mt-", col.name = "percent.mito")
      } else {
        DefaultAssay(seuratSeqGeq) <- 'RNA'
        seuratSeqGeq <- PercentageFeatureSet(seuratSeqGeq, pattern = "^MT-", col.name = "percent.mito")
      }
    }
  } else {
    if (checkMitoGenes){
      hu.mito.genes <- grep(pattern = "^MT-", x = rownames(x = seuratSeqGeq@assays[["RNA"]]), value = TRUE)
      if (length(hu.mito.genes) < 1)
      { # Possibly mouse
        mo.mito.genes <- grep(pattern = "^mt-", x = rownames(x = seuratSeqGeq@assays[["RNA"]]), value = TRUE)
        DefaultAssay(seuratSeqGeq) <- 'RNA'
        seuratSeqGeq <- PercentageFeatureSet(seuratSeqGeq, pattern = "^mt-", col.name = "percent.mito")
      } else {
        DefaultAssay(seuratSeqGeq) <- 'RNA'
        seuratSeqGeq <- PercentageFeatureSet(seuratSeqGeq, pattern = "^MT-", col.name = "percent.mito")
      }
    }
  }

  #%%%%%%%%%%%%%%%%%%%%%%%#
  print(summary(seuratSeqGeq@meta.data))
  #%%%%%%%%%%%%%%%%%%%%%%%#

  if (filterCells)
  {
    subsetList <- c("nGene")
    SG_LOW_THRESHOLD_VALUE <- quantile(seuratSeqGeq@meta.data$nFeature_RNA, 0.07)
    SG_HIGH_THRESHOLD_VALUE <- quantile(seuratSeqGeq@meta.data$nFeature_RNA, 0.99)
    nCountLowerQuartile <- quantile(seuratSeqGeq@meta.data$nCount_RNA, 0.2)
    nCountUpperQuartile <- quantile(seuratSeqGeq@meta.data$nCount_RNA, 0.80)
    seuratSeqGeq <- subset(seuratSeqGeq, nFeature_RNA > SG_LOW_THRESHOLD_VALUE & nFeature_RNA < SG_HIGH_THRESHOLD_VALUE)
  }

  # Log normalize data, add scaling factor
  if (logNormalizeBool)
  {
    scalingFactor = 10000
    adjustScalingFactor = FALSE
    if (adjustScalingFactor)
    {
      scalingFactor = SG_SCALING_FACTOR
    }
    if(runWeightedNearestNeighbors){
      DefaultAssay(seuratSeqGeq) <- 'RNA'
      seuratSeqGeq <- NormalizeData(object = seuratSeqGeq, normalization.method = "LogNormalize",
                                    scale.factor = scalingFactor)

      DefaultAssay(seuratSeqGeq) <- 'ADT'
      seuratSeqGeq <- NormalizeData(seuratSeqGeq, normalization.method = 'CLR', margin = 2)
    }
    seuratSeqGeq <- NormalizeData(object = seuratSeqGeq, normalization.method = "LogNormalize",
                                  scale.factor = scalingFactor)
    exportNormData = FALSE
    if (exportNormData)
    {
      exportNormalizedData <- as.data.frame(seuratSeqGeq@scale.data)
    }
  }
  # Find variable genes
  findVarGenes = TRUE
  if (findVarGenes)
  {
    DefaultAssay(seuratSeqGeq) <- 'RNA'
    # Find variables genes and add to Seurat object.
    seuratSeqGeq <- FindVariableFeatures(seuratSeqGeq, selection.method = "vst")
    # Declare minimum genes to find, default is 10, can be more.
    minimumVarGenes = 10
    minVarGeneBool = FALSE
    if (minVarGeneBool)
    {
      minimumVarGenes = SG_MINIMUM_VAR_GENES
    }

    if (length(seuratSeqGeq@assays[["RNA"]]@meta.data$var.features) < minimumVarGenes)
    {
      seuratSeqGeq <- FindVariableFeatures(seuratSeqGeq, selection.method = "vst")
    }
  }

  if(length(hu.mito.genes > 0 )){
    qcPlot <- VlnPlot(seuratSeqGeq, features = c("nFeature_RNA", "nCount_RNA", "percent.mito"), ncol = 3, pt.size = 0.2)
  } else if(length(mo.mito.genes > 0 )){
    qcPlot <- VlnPlot(seuratSeqGeq, features = c("nFeature_RNA", "nCount_RNA", "percent.mito"), ncol = 3, pt.size = 0.2)
  } else {
    qcPlot <- VlnPlot(seuratSeqGeq, features = c("nFeature_RNA", "nCount_RNA"), ncol = 2, pt.size = 0.2)
  }
  ggsave(file=paste0("ViolinQCPlot_", runID, ".png"), device= "png", plot= qcPlot, height= 900, width= 1500, units= "px", dpi=300, bg = "transparent")



  ##Scale the data
  if (logNormalizeBool)
  {
    if(runWeightedNearestNeighbors){
      DefaultAssay(seuratSeqGeq) <- 'RNA'
      rna.genes <- rownames(seuratSeqGeq[["RNA"]])
      seuratSeqGeq <- ScaleData(seuratSeqGeq, features = rna.genes)

      DefaultAssay(seuratSeqGeq) <- 'ADT'
      adt.features <- rownames(seuratSeqGeq[["ADT"]])
      seuratSeqGeq <- ScaleData(seuratSeqGeq, features = adt.features)
    }else{
      all.genes <- rownames(seuratSeqGeq)
      seuratSeqGeq <- ScaleData(seuratSeqGeq, features = all.genes)
    }
  }

  if (scTransform) {
    if(length(hu.mito.genes > 0 )){
      seuratSeqGeq <- SCTransform(seuratSeqGeq, variable.features.n = 3000,
                                  variable.features.rv.th = 1.3, seed.use = 777,
                                  verbose = TRUE, do.scale = F, do.center = T,
                                  vars.to.regress = "percent.mito")
    } else if (length(mo.mito.genes > 0)){
      seuratSeqGeq <- SCTransform(seuratSeqGeq, variable.features.n = 3000,
                                  variable.features.rv.th = 1.3, seed.use = 777,
                                  verbose = TRUE, do.scale = F, do.center = T,
                                  vars.to.regress = "percent.mito")
    } else {
      seuratSeqGeq <- SCTransform(seuratSeqGeq, variable.features.n = 3000,
                                  variable.features.rv.th = 1.3, seed.use = 777,
                                  verbose = TRUE, do.scale = F, do.center = T)
    }
    # also need to scaleData() below for the ADT assay slot
       if(runWeightedNearestNeighbors){
        DefaultAssay(seuratSeqGeq) <- 'ADT'
        adt.features <- rownames(seuratSeqGeq[["ADT"]])
        seuratSeqGeq <- ScaleData(seuratSeqGeq, features = adt.features)
       }
  }

  regressOutVars = TRUE
  if (regressOutVars)
  {
    # These cell cycle gene sets are available in SeqGeq Extras folder.
    # create a gene set without cell cycle genes instead of performing this step

    # Seurat automatically loads gene sets from Tirosh et al.
    # these can be used to regress out cell cycle genes
    # s.genes <- cc.genes$s.genes
    # g2m.genes <- cc.genes$g2m.genes
    # seuratSeqGeq <- CellCycleScoring(seuratSeqGeq, s.features = s.genes, g2m.features = g2m.genes, set.ident = FALSE)
    # seuratSeqGeq <- ScaleData(object = seuratSeqGeq, vars.to.regress = c("S.Score","G2M.Score"))

    if(length(hu.mito.genes > 0 )){
      seuratSeqGeq <- ScaleData(object = seuratSeqGeq, vars.to.regress = "percent.mito")
    } else if(length(mo.mito.genes > 0 )){
      seuratSeqGeq <- ScaleData(object = seuratSeqGeq, vars.to.regress = "percent.mito")
    } else {
      seuratSeqGeq <- ScaleData(object = seuratSeqGeq, vars.to.regress = NULL)
    }
  }
  #%%%%%%%%%%%%%%%%%%%%%%%#
  ## Run DimRedux ##
  #%%%%%%%%%%%%%%%%%%%%%%%#
  runDimRedux = TRUE
  runPCA = TRUE
  if (runPCA)
  {
    if(runWeightedNearestNeighbors){
      adtlength <- length(adtParNames)
      apcaDim <- adtlength - 2
      if(scTransform){
        DefaultAssay(seuratSeqGeq) <- 'SCT'
        seuratSeqGeq <- RunPCA(seuratSeqGeq, reduction.name = 'pca', npcs = 30, redcution.key = "rnaPC")
        DefaultAssay(seuratSeqGeq) <- 'ADT'
        # we will use all ADT features for dimensional reduction
        # we set a dimensional reduction name to avoid overwriting the
        VariableFeatures(seuratSeqGeq) <- rownames(seuratSeqGeq[["ADT"]])
        seuratSeqGeq <- RunPCA(seuratSeqGeq, reduction.name = 'apca', npcs = apcaDim, reduction.key = "adtPC")
        exportPCA1 <- seuratSeqGeq@reductions[["pca"]]@cell.embeddings
        exportPCA2 <- seuratSeqGeq@reductions[["apca"]]@cell.embeddings
        exportPCA <- cbind(exportPCA1, exportPCA2)
        writeLines(" Finished: Calculating wnnPCA components")
      }else{ # logTransform
        DefaultAssay(seuratSeqGeq) <- 'RNA'
        seuratSeqGeq <- RunPCA(seuratSeqGeq, reduction.name = 'pca', npcs = 30, redcution.key = "rnaPC")
        DefaultAssay(seuratSeqGeq) <- 'ADT'
        # we will use all ADT features for dimensional reduction
        # we set a dimensional reduction name to avoid overwriting the
        VariableFeatures(seuratSeqGeq) <- rownames(seuratSeqGeq[["ADT"]])
        seuratSeqGeq <- RunPCA(seuratSeqGeq, reduction.name = 'apca', npcs = apcaDim, reduction.key = "adtPC")
        exportPCA1 <- seuratSeqGeq@reductions[["pca"]]@cell.embeddings
        exportPCA2 <- seuratSeqGeq@reductions[["apca"]]@cell.embeddings
        exportPCA <- cbind(exportPCA1, exportPCA2)
        writeLines(" Finished: Calculating wnnPCA components")
      }

    }else{

      if (scTransform){
        seuratSeqGeq <- RunPCA(seuratSeqGeq)
      } else {
        seuratSeqGeq <- FindVariableFeatures(seuratSeqGeq, selection.method = "vst")
        seuratSeqGeq <- RunPCA(seuratSeqGeq, features = VariableFeatures(seuratSeqGeq))
      }
      exportPCA <- seuratSeqGeq@reductions[["pca"]]@cell.embeddings
      writeLines(" Finished: Calculating PCA components")
    }
    ### NOTE: Getting an Error - "You're computing too large a percentage of total singular values,
    ##          use a standard svd instead."
    ## Quoting the Satija Lab, "We use an approximate algorithm for PCA which doesn't provide much
    ##          benefit if the number of features is close to the number of PCs requested. It is
    ##          usually safe to ignore this message."
    ## <https://github.com/satijalab/seurat/issues/1249>
  }

#' Identify multimodal neighbors. These will be stored in the neighbors slot,
#' and can be accessed using seuratSeqGeq[['weighted.nn']]
#' The WNN graph can be accessed at seuratSeqGeq[["wknn"]],
#' Cell-specific modality weights can be accessed at seuratSeqGeq$RNA.weight

if(runWeightedNearestNeighbors){
  adtlength <- length(seuratSeqGeq@reductions[['apca']])
  rnalength <- length(seuratSeqGeq@reductions[['pca']])
    seuratSeqGeq <- FindMultiModalNeighbors(
      seuratSeqGeq, reduction.list = list("pca", "apca"),
      dims.list = list(1:rnalength, 1:adtlength), modality.weight.name = "RNA.weight"
    )
}

} else {
  writeLines("Starting Batch Correction")
  # Seurat exports as PIR (Parameters in Rows)
  # Transpose the CSV so that the column 'SampleID' can be used for formatting prior to making a Seurat Object
  newDF <- data.frame(t(importedCSVfile))
  splitDF <- split(newDF, newDF$SampleID)
  for(i in seq_along(splitDF)) {
    splitDF[[i]] <- subset(splitDF[[i]], select = -c(SampleID))
  }
  # Now transpose them to params in cloumns for Seurat
  for(i in seq_along(splitDF)){
    splitDF[[i]] <- t(splitDF[[i]])
  }
  ###################################
  ## Create List of Seurat Objects ##
  ###################################
  if (runQC){ #TODO check if SampleID row should be exluded from this calculation of concat file
    if (minCellCutoffBool) {
      minCellPercentVal <- cellPercent
      minCellCutoff <- nrow(importedCSVfile) * minCellPercentVal
    } else { minCellCutoff <- cellPercent }

    if(minGeneCutoffBool) {
      minGeneCutoff <- ncol(importedCSVfile) * minGenePercent
    } else { minGeneCutoff <- minGenePercent }

    if (logNormalizeBool){
      logNormalizeData <- "logNormalize"
    }
    output <- vector("list", length(splitDF))
    for (i in seq_along(splitDF)) {
      output[[i]] <- CreateSeuratObject(splitDF[[i]],
                                        min.cells = minCellCutoff,
                                        min.features = minGeneCutoff)
      #, project = namesToAdd[i])#TODO can we get hashmap or names of concat files?
    }
    rm(importedCSVfile)
  }else {
    # No QC, just create Seurat object
    output <- vector("list", length(splitDF))
    for (i in seq_along(splitDF)) {            # 2. sequence
      output[[i]] <- CreateSeuratObject(splitDF[[i]]) #, project = namesToAdd[i])      # 3. body
    }
    rm(importedCSVfile)
  }

  checkMitoGenes = TRUE
  if (yourVersion < 5.0) {
    if (checkMitoGenes){
      for (i in seq_along(output)){
        hu.mito.genes <- grep(pattern = "^MT", x = rownames(x = output[[i]]@assays[["RNA"]]@data), value = TRUE)
      }
      if (length(hu.mito.genes) < 1) { # Possibly mouse
        for (i in seq_along(output)){
          mo.mito.genes <- grep(pattern = "^mt", x = rownames(x = output[[i]]@assays[["RNA"]]@data), value = TRUE)
          # DefaultAssay(seuratSeqGeq) <- 'RNA'
          output[[i]] <- PercentageFeatureSet(object = output[[i]], pattern = "^mt-", col.name = "percent.mito")
        }
      } else {
        for (i in seq_along(output)){
          # DefaultAssay(seuratSeqGeq) <- 'RNA'
          output[[i]] <- PercentageFeatureSet(object = output[[i]], pattern = "^MT-", col.name = "percent.mito")
        }
      }
    }
  } else {
    if (checkMitoGenes){
      for (i in seq_along(output)){
        hu.mito.genes <- grep(pattern = "^MT", x = rownames(x = output[[i]]@assays[["RNA"]]), value = TRUE)
      }
      if (length(hu.mito.genes) < 1) { # Possibly mouse
        for (i in seq_along(output)){
          mo.mito.genes <- grep(pattern = "^mt", x = rownames(x = output[[i]]@assays[["RNA"]]), value = TRUE)
          # DefaultAssay(seuratSeqGeq) <- 'RNA'
          output[[i]] <- PercentageFeatureSet(object = output[[i]], pattern = "^mt-", col.name = "percent.mito")
        }
      } else {
        for (i in seq_along(output)){
          # DefaultAssay(seuratSeqGeq) <- 'RNA'
          output[[i]] <- PercentageFeatureSet(object = output[[i]], pattern = "^MT-", col.name = "percent.mito")
        }
      }
    }
  }

  if (filterCells){
    subsetList = c("nGene")

    for (i in seq_along(output)){
      SG_LOW_THRESHOLD_VALUE <- quantile(output[[i]]@meta.data$nFeature_RNA, 0.1)
      SG_HIGH_THRESHOLD_VALUE <- quantile(output[[i]]@meta.data$nFeature_RNA, 0.99)
      output[[i]] <- subset(output[[i]], nFeature_RNA > SG_LOW_THRESHOLD_VALUE & nFeature_RNA < SG_HIGH_THRESHOLD_VALUE)
    }
  }

  # Log normalize data, add scaling factor
  if (logNormalizeBool){
    scalingFactor = 10000
    adjustScalingFactor = FALSE
    if (adjustScalingFactor){
      scalingFactor = SG_SCALING_FACTOR
    }
    # Normalize data to scaling factor and method, def: LogNorm. TODO: Check for future methods
    for (i in seq_along(output)){
      output[[i]] <- NormalizeData(object = output[[i]], normalization.method = "LogNormalize",
                                   scale.factor = scalingFactor)
    }
    exportNormData = FALSE
    outputNormDataList <- vector("list", length(output))
    if (exportNormData){
      for (i in seq_along(outputNormDataList)){
        outputNormDataList[[i]] <- as.data.frame(outputNormDataList[[i]]@scale.data)}
    }
    for (i in seq_along(output)){
      output[[i]] <- FindVariableFeatures(output[[i]], selection.method='vst', clip.max='auto', nfeatures=2500 )
    }

  }

  if (scTransform) {
    # Shaowen Jiang, NYU - Deal with variable features in a more elegent way.
    ## Problem: If we set this to the default 3000, it throws an error if < 3000 features are present in the data.frame
    ## What we were doing:
    # seuratSeqGeq <- SCTransform(seuratSeqGeq, variable.features.n = NULL, variable.features.rv.th = 1.3, seed.use = 777, verbose = TRUE)
    ## Defaults:
    if(length(hu.mito.genes > 0 )){
      for (i in seq_along(output)){
        output[[i]] <- SCTransform(output[[i]], variable.features.n = 3000,
                                   variable.features.rv.th = 1.3, seed.use = 777,
                                   verbose = TRUE, do.scale = F, do.center = T,
                                   vars.to.regress = "percent.mito")
      }
    } else if (length(mo.mito.genes > 0)){
      for (i in seq_along(output)){
        output[[i]] <- SCTransform(output[[i]], variable.features.n = 3000,
                                   variable.features.rv.th = 1.3, seed.use = 777,
                                   verbose = TRUE, do.scale = F, do.center = T,
                                   vars.to.regress = "percent.mito")
      }
    } else {
      for (i in seq_along(output)){
        output[[i]] <- SCTransform(output[[i]], variable.features.n = 3000,
                                   variable.features.rv.th = 1.3, seed.use = 777,
                                   verbose = TRUE, do.scale = F, do.center = T)
      }
    }
    # Q: How does this compare?: (changed to VariableFeatures(seuratSeqGeq)) A: No difference(?), strange
    # seuratSeqGeq <- SCTransform(seuratSeqGeq, variable.features.n = VariableFeatures(seuratSeqGeq), variable.features.rv.th = 1.3, seed.use = 777, verbose = TRUE)
  }

  # Next, select features for downstream integration, and run PCA on each object in the list,
  # which is required for running the reciprocal PCA workflow.
  myFeatures <- SelectIntegrationFeatures(object.list = output, nfeatures = 2500)

  if(scTransform){
    output <- PrepSCTIntegration(output, anchor.features = myFeatures, verbose = FALSE)
  }

  output <- lapply(X = output, FUN = function(x) {
    x <- ScaleData(x, features = myFeatures, verbose = FALSE)
    x <- RunPCA(x, features = myFeatures, verbose = FALSE)
  })



  # Next, identify anchors and integrate the datasets. Commands are identical to the standard workflow, but make sure
  # to set reduction = 'rpca':
  if(scTransform){
    anchors <- FindIntegrationAnchors(object.list = output, reference = NULL, reduction = "rpca",
                                      dims = 1:50, normalization.method = "SCT", anchor.features = myFeatures)
  } else {
    anchors <- FindIntegrationAnchors(object.list = output, reference = NULL, reduction = "rpca",
                                      dims = 1:50)
  }

  output.batchCorrected <- IntegrateData(anchorset = anchors, dims = 1:50) # Consider using "preserve.order=TRUE"
  # Now proceed with downstream analysis (i.e. visualization, clustering) on the integrated dataset.
  # Commands are identical to the standard workflow.
  output.batchCorrected <- ScaleData(output.batchCorrected, verbose = FALSE)

  if (scTransform){
    output.batchCorrected <- PrepSCTFindMarkers(output.batchCorrected, assay = "SCT",verbose = TRUE)
    output.batchCorrected <- RunPCA(output.batchCorrected)
  } else {
    output.batchCorrected <- RunPCA(output.batchCorrected, features = VariableFeatures(output.batchCorrected))
  }

  exportPCA <- output.batchCorrected@reductions[["pca"]]@cell.embeddings
  writeLines(" Finished: Calculating PCA components")

  seuratSeqGeq <- output.batchCorrected

}

## Find Clusters ##
if(runWeightedNearestNeighbors){
  seuratSeqGeq <- FindClusters(seuratSeqGeq, graph.name = "wsnn", algorithm = 3, resolution = rezza, verbose = FALSE)

  seuratSeqGeq$seurat_clusters <- as.factor(as.numeric(as.character(seuratSeqGeq$seurat_clusters)) + 1)

  num.clust <-  length(levels(seuratSeqGeq@meta.data[["seurat_clusters"]]))

  exportKNNClusters0 <- as.data.frame(seuratSeqGeq@meta.data[["seurat_clusters"]])
  exportKNNClusters1 <- as.matrix(exportKNNClusters0)

  exportKNNClusters2 <- as.numeric(exportKNNClusters1) # + 1
  exportKNNClusters <- as.data.frame(exportKNNClusters2)
  colnames(exportKNNClusters) <- c("Seurat_Clusters")

  write.table(num.clust, paste0("Estimated_k_", runID), row.names = F, col.names = F)
}else{
  seuratSeqGeq <- FindNeighbors(seuratSeqGeq, dims = 1:10)
  seuratSeqGeq <- FindClusters(seuratSeqGeq, resolution = rezza)

  seuratSeqGeq$seurat_clusters <- as.factor(as.numeric(as.character(seuratSeqGeq$seurat_clusters)) + 1)

  num.clust <-  length(levels(seuratSeqGeq@meta.data[["seurat_clusters"]]))

  exportKNNClusters0 <- as.data.frame(seuratSeqGeq@meta.data[["seurat_clusters"]])
  exportKNNClusters1 <- as.matrix(exportKNNClusters0)

  exportKNNClusters2 <- as.numeric(exportKNNClusters1) # + 1
  exportKNNClusters <- as.data.frame(exportKNNClusters2)
  colnames(exportKNNClusters) <- c("Seurat_Clusters")

  write.table(num.clust, paste0("Estimated_k_", runID), row.names = F, col.names = F)
}

if (runTSNE)
{
changePerplexity = SG_CHANGE_PERPLEXITY
    perplexityVal = 20
    if (changePerplexity && is.numeric(perplexityVal))
    {
        perplexityVal = SG_TSNE_PERPLEXITY
    }

    # Run rTSNE in Seurat wrapper, use deterministic, assign perplexty
    seuratSeqGeq <- RunTSNE(object = seuratSeqGeq, check_duplicates = FALSE, seed.use = 1, perplexity = perplexityVal, dims.use = 1 : 10, do.fast = TRUE)
    exportTSNE <- seuratSeqGeq@reductions[["tsne"]]@cell.embeddings

    colnames(exportTSNE) <- c("tSNE_X", "tSNE_Y");
}

if (runUMAP)
{
  if(runWeightedNearestNeighbors){
    seuratSeqGeq <- RunUMAP(seuratSeqGeq, nn.name = "weighted.nn", reduction.name = "wnn.umap", reduction.key = "wnnUMAP_")
    exportUMAP <- seuratSeqGeq@reductions[["wnn.umap"]]@cell.embeddings
    colnames(exportUMAP) <- c("wnnUMAP_X", "wnnUMAP_Y");
  }else {
    # Run rUMAP in Seurat wrapper, use deterministic, assign perplexity
    # seuratSeqGeq <- RunTSNE(object = seuratSeqGeq, check_duplicates = FALSE, seed.use = 1, perplexity = perplexityVal, dims.use = 1 : 10, do.fast = TRUE)
    seuratSeqGeq <- RunUMAP(seuratSeqGeq, dims = 1:10, verbose = FALSE)
    exportUMAP <- seuratSeqGeq@reductions[["umap"]]@cell.embeddings
    colnames(exportUMAP) <- c("UMAP_X", "UMAP_Y");
  }
}

#%%%%%%%%%%%%%%%%%%%%%%%#
#%%%%%%%%%%%%%%%%%%%%%%%#
#%%%%%%%%%%%%%%%%%%%%%%%#
## Clustering ##
#%%%%%%%%%%%%%%%%%%%%%%%#

    findClusterGenes = TRUE
    if (findClusterGenes)
    {
        writeLines(" Start: Clustering and analyzing differentially expressed genes per cluster, may take a little while...")
        findClusterMethod = "negbinom"
        changeFindClusterMethod = TRUE
        if (changeFindClusterMethod)
        {
            ## Only allow "bimod", "roc", "t", , "tobit", "poisson", "negbinom", default is negbinom
            findClusterMethod = "SG_FIND_CLUSTER_METHOD"
        }

      if(scTransform){
        DefaultAssay(seuratSeqGeq) <- 'SCT'
      }else{
        DefaultAssay(seuratSeqGeq) <- 'RNA'
      }

        ## Make sure the cluster numbers are sensible. (1... n, not zero indexed)
        seuratSeqGeq@meta.data$seurat_clusters <- as.factor(as.numeric(seuratSeqGeq@meta.data$seurat_clusters))

        ## Run test to find top markers for each cluster found in KNN
      if(runBatchCorrection){
        seuratSeqGeq.markers <- FindAllMarkers(object = seuratSeqGeq, only.pos = F, min.pct = 0.25, test.use = "wilcox", verbose = TRUE, random.seed = FALSE)
      }else{
        seuratSeqGeq.markers <- FindAllMarkers(object = seuratSeqGeq, only.pos = F, min.pct = 0.25, test.use = findClusterMethod, verbose = TRUE, random.seed = FALSE)
        seuratSeqGeq.markers.RNA <- FindAllMarkers(object = seuratSeqGeq, only.pos = F, min.pct = 0.25, test.use = findClusterMethod, verbose = TRUE, random.seed = FALSE)
      }

      if(runWeightedNearestNeighbors){
        DefaultAssay(seuratSeqGeq) <- 'ADT'
        seuratSeqGeq.markers.ADT <- FindAllMarkers(object = seuratSeqGeq, only.pos = F, min.pct = 0.25, test.use = findClusterMethod, verbose = TRUE, random.seed = FALSE)

        seuratSeqGeq.markers.X <- rbind(seuratSeqGeq.markers, seuratSeqGeq.markers.ADT)
        seuratSeqGeq.markers <- seuratSeqGeq.markers.X
      }

        seuratSeqGeq.markers$cluster <- as.numeric(seuratSeqGeq.markers$cluster)
        seuratSeqGeq@active.ident <- as.factor(as.numeric(seuratSeqGeq@active.ident))

        col_idx <- grep("gene", names(seuratSeqGeq.markers))
        seuratSeqGeq.markers <- seuratSeqGeq.markers[, c(col_idx, (1 : ncol(seuratSeqGeq.markers))[- col_idx])]

        # Put top genes before clusters
        wheresymbol <- which(names(seuratSeqGeq.markers) == "gene")
        whereCluster <- which(names(seuratSeqGeq.markers) == "cluster")

        # Create Marker Gene Export
        top10 <- seuratSeqGeq.markers %>% group_by(cluster) %>% top_n(10, avg_log2FC)
        upFC <- top10$avg_log2FC
        qVal <- top10$p_val_adj
        # upFC <- top10$avg_logFC
        feat <- top10$gene
        clust <- top10$cluster
        top10feats <- cbind(feat, upFC, qVal, clust)
        nam <- paste("Seurat_Marker_Features", sep = "_")
        assign(nam, top10feats)

        namGS <- paste0("/Seurat_Marker_Features", "_", runID, ".csv")
        fGSnam <- paste0(outfileDirectory,namGS)
        write.table("[GeneSet]", file = fGSnam, append = F, quote = F, row.names = F, col.names = F )
        write.table("$SchemaVersion,1", file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
        write.table(paste0("$GeneSetName,Seurat Marker Features ", "RunID: ", runID), file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
        write.table(paste0("$GeneSetDescription,Generated in the Seurat analysis pipeline. Fold-Change > ",min(upFC),"; q-Val (FDR) &lt; ", max(qVal)), file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
        write.table(" ", file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
        write.table("[Genes],Log2 Fold-Change,q-Value,Cluster", file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
        write.table(top10feats, file = fGSnam, sep = ',', append = T, quote = F, row.names = F, col.names = F , na="")
        ##%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%##
        ##%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%##
        # Write out DEG sets #
        ##%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%##
        ## UP ##
        for (n in 1:num.clust){
          upDEGr <- subset(seuratSeqGeq.markers, (seuratSeqGeq.markers$cluster == (n)))
          upDEGr <- subset(upDEGr, upDEGr$avg_log2FC > 0)
          qVal <- upDEGr$p_val_adj
          upFC <- upDEGr$avg_log2FC
          feat <- upDEGr$gene
          upDEG <- cbind(feat, upFC, qVal)
          nam <- paste("Up_in_Cluster", n, sep = "_")
          assign(nam, upDEG)
          namGS <- paste0("/Up_in_Cluster", "_", n, "_", runID, ".csv")
          fGSnam <- paste0(outfileDirectory,namGS)
          write.table("[GeneSet]", file = fGSnam, append = F, quote = F, row.names = F, col.names = F )
          write.table("$SchemaVersion,1", file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
          write.table(paste0("$GeneSetName,Up in Cluster ", n, " RunID: ", runID), file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
          write.table(paste0("$GeneSetDescription,Generated in the Seurat analysis pipeline. Fold-Change > ",min(upFC),"; q-Val (FDR) &lt; ", max(qVal)), file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
          write.table(" ", file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
          write.table("[Genes],Log2 Fold-Change,q-Value", file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
          write.table(upDEG, file = fGSnam, sep = ',', append = T, quote = F, row.names = F, col.names = F , na="")
        }
            ## DOWN ##
        for (n in 1:num.clust){
          downDEGr <- subset(seuratSeqGeq.markers, (seuratSeqGeq.markers$cluster == (n)))
          downDEGr <- subset(downDEGr, downDEGr$avg_log2FC < 0)
          qVal <- downDEGr$p_val_adj
          downFC <- downDEGr$avg_log2FC
          feat <- downDEGr$gene
          downDEG <- cbind(feat, downFC, qVal)
          nam <- paste("Down_in_Cluster", n, sep = "_")
          assign(nam, downDEG)
          namGS <- paste0("/Down_in_Cluster", "_", n, "_", runID, ".csv")
          fGSnam <- paste0(outfileDirectory,namGS)
          write.table("[GeneSet]", file = fGSnam, append = F, quote = F, row.names = F, col.names = F )
          write.table("$SchemaVersion,1", file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
          write.table(paste0("$GeneSetName,Down in Cluster ", n, " RunID: ", runID), file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
          write.table(paste0("$GeneSetDescription,Generated in the Seurat analysis pipeline. Fold-Change &lt; ",min(downFC),"; q-Val (FDR) &lt; ", max(qVal)), file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
          write.table(" ", file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
          write.table("[Genes],Log2 Fold-Change,q-Value", file = fGSnam, append = T, quote = F, row.names = F, col.names = F )
          write.table(downDEG, file = fGSnam, sep = ',', append = T, quote = F, row.names = F, col.names = F , na="")
        }
          write.csv(seuratSeqGeq.markers, file = paste0(findClusterMethod, ".Summary.csv"), quote = FALSE, row.names = FALSE)
    }

if (runTSNE && !runUMAP) {
    exportDimReduxClustering = cbind(exportPCA, exportTSNE, exportKNNClusters)
} else if (runUMAP && !runTSNE) {
    exportDimReduxClustering = cbind(exportPCA, exportUMAP, exportKNNClusters)
} else if (runTSNE && runUMAP) {
    exportDimReduxClustering = cbind(exportPCA, exportTSNE, exportUMAP, exportKNNClusters)
} else {
    exportDimReduxClustering = cbind(exportPCA, exportKNNClusters)
}
    rowNamesDimRedux <- as.data.frame(row.names(exportDimReduxClustering))
    colnames(rowNamesDimRedux) <- "CellId"
    exportDimReduxClustering <- cbind(rowNamesDimRedux, exportDimReduxClustering)
    is.num <- sapply(exportDimReduxClustering, is.numeric)
    exportDimReduxClustering[is.num] <- lapply(exportDimReduxClustering[is.num], round, 2)
    colnames(exportDimReduxClustering) <- paste(colnames(exportDimReduxClustering), runID, sep = "_")
    colnames(exportDimReduxClustering)[1] <- "CellId"
    exportDimReduxClustering[, 1] <- gsub("AppID_*", "", rownames(exportDimReduxClustering))
    write.csv(exportDimReduxClustering, file = paste0("PCA_tSNE_Clustering_", runID, ".csv"), row.names = FALSE, quote = FALSE)
##%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%##
##%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%##
#%%%%%%%%%%%%%%%%%%%%%%%#
## Classification ##
#%%%%%%%%%%%%%%%%%%%%%%%#
## NOTE: This set of parameters is very tricky to find - pMonocytes could be better identified by BCL2A1.
#        The logic for this is very basic - and it requires that a feature be absolutely uniqe to the population of interest.
#        Would benefit from some more advanced boolean logic down the line... (decision trees)
#        Kind of a cool idea none the less.
     if (classifyClusters) {

            if(model == "huPBMC") {
              ClassGeneSet = c("IL7R", "LYZ", "MS4A1", "CD8B", "CXCL16", "KLRF1", "FCER1A", "PPBP", "GZMK", "FOXP3")
              standard = 1.5
            } else if (model == "mmLiver") {
              ClassGeneSet = c("Cd19","Gzma","Cd8a","Cd38","Spp1","F13a1","Cxcr6","Clec4d","Flt3") # ,"","")
              standard = 1.5
            } else if (model == "mmLN_Spleen") {
              ClassGeneSet = c("Gzma","Ms4a1","Cd8a","Slfn1","Fcgr3","Cxcl16","Vcam1","Retnlg")
              standard = 1.1
            } else if (model == "mmBrain") {
                ClassGeneSet = c("Tle4","Gpr88","Gad2","Cux2","Slc1a2","Sst","Plp1","Pdgfra","Cck","Adam33","Garnl3","Vip")
                standard = 1.2
            } else if (model == "huCD45+Skin") {
                ClassGeneSet = c("CD14","LAMP3","FCER1A","IL7R","PRF1","IRF8","CPA3","GZMH","GZMK","FOXP3","MS4A1", "CD69")
                standard = 1.5
            } else if (model == "huBM"){
                ClassGeneSet = c("MME","CD19","DNAJB9","ALAS2","PMP22","SMIM24","PDLIM1","CRHBP","IL3RA","FGR","GZMA","CD2","GNLY")
                standard = 1.5
            }

            if (scTransform){
              ClassGeneSet_rd = Reduce(intersect, list(ClassGeneSet, seuratSeqGeq@assays$SCT$data@Dimnames[[1]]))
            } else {
              ClassGeneSet_rd = Reduce(intersect, list(ClassGeneSet, seuratSeqGeq@assays$RNA@layers$counts@Dimnames[[1]]))
            }


         if(length(ClassGeneSet_rd) > 2){
             geneSize = length(ClassGeneSet)
             foo <- vector(mode="list", length=(geneSize))
             if(model == "huPBMC") {
                 names(foo) <- c(
                 "IL7R",
                 "LYZ",
                 "MS4A1",
                 "CD8B",
                 "CXCL16",
                 "KLRF1",
                 "FCER1A",
                 "PPBP",
                 "GZMK",
                 "FOXP3")
                 foo[[1]] <- "IL7R+ T4 cells";
                 foo[[2]] <- "LYZ+ cMonocytes";
                 foo[[3]] <- "MS4A1+ B cells";
                 foo[[4]] <- "CD8B+ T8n cells";
                 foo[[5]] <- "CXCL16+ pMonocytes";
                 foo[[6]] <- "KLRF1+ NK cells";
                 foo[[7]] <- "FCER1A+ Dendritic cells";
                 foo[[8]] <- "PPBP+ Megakaryocytes";
                 foo[[9]] <- "GZMK+ T8m cells";
                 foo[[10]] <- "FOXP3+ Treg cells";
             } else if (model == "mmLiver") {
                 names(foo) <- c(
                 "Cd19",
                 "Gzma",
                 "Cd8a",
                 "Cd38",
                 "Spp1",
                 "F13a1",
                 "Cxcr6",
                 "Clec4d",
                 "Flt3")
                 # ,
                 # "")
                 foo[[1]] <- "Cd19+ B cells";
                 foo[[2]] <- "Gzma+ NK cells";
                 foo[[3]] <- "Cd8a+ T8 cells";
                 foo[[4]] <- "Cd38+ Hepatic Stellate cells";
                 foo[[5]] <- "Spp1+ LSEndothelial cells";
                 foo[[6]] <- "F13a1+ Hepatocytes";
                 foo[[7]] <- "Cxcr6+ T4 cells";
                 foo[[8]] <- "Clec4d+ Kupffer cells";
                 foo[[9]] <- "Flt3+ Hepatic Stellate cells";
                 # foo[[10]] <- "+  cells";
             } else if (model == "mmLN_Spleen") {
                 names(foo) <- c(
                 "Gzma",
                 "Ms4a1",
                 "Cd8a",
                 "Slfn1",
                 "Fcgr3",
                 "Cxcl16",
                 "Vcam1",
                 "Retnlg")
                 foo[[1]] <- "Gzma+ NK cells";
                 foo[[2]] <- "Ms4a1+ B cells";
                 foo[[3]] <- "Cd8a+ T8 cells";
                 foo[[4]] <- "Slfn1+ T4 cells";
                 foo[[5]] <- "Fcgr3+ cMonocytes";
                 foo[[6]] <- "Cxcl16+ pMonocytes";
                 foo[[7]] <- "Vcam+ Endothelial cells";
                 foo[[8]] <- "Retnlg+ Macrophage cells";
                 # foo[[9]] <- "";
             } else if (model == "mmBrain") {
                 names(foo) <- c(
                 "Tle4",
                 # "Nrgn",
                 "Gpr88",
                 "Gad2",
                 "Cux2",
                 "Slc1a2",
                 "Sst",
                 "Plp1",
                 "Pdgfra",
                 "Cck",
                 "Adam33",
                 "Garnl3",
                 "Vip")
                 foo[[1]] <- "Tle4+ Corticofugal Projection Neurons";
                 # foo[[2]] <- "Nrgn+ Neurons";
                 foo[[2]] <- "Gpr88+ Striatum";
                 foo[[3]] <- "Gad2+ Interneurons";
                 foo[[4]] <- "Cux2+ Excitatory Neurons";
                 foo[[5]] <- "Slc1a2+ Astrocytes";
                 foo[[6]] <- "Sst+ Interneurons";
                 foo[[7]] <- "Plp1+ Oligodendrocytes";
                 foo[[8]] <- "Pdgfra+ preOligodendrocytes";
                 foo[[9]] <- "Cck+ Excitatory Neurons";
                 foo[[10]] <- "Adam33+ Excitatory Neurons";
                 foo[[11]] <- "Garnl3+ Maturing Neurons";
                 foo[[12]] <- "Vip+ Interneurons";
                 # foo[[9]] <- "";
             } else if (model == "huCD45+Skin") {
                 names(foo) <- c(
                 "CD14",
                 "LAMP3",
                 "FCER1A",
                 "IL7R",
                 "PRF1",
                 "IRF8",
                 "CPA3",
                 "GZMH",
                 "GZMK",
                 "FOXP3",
                 "MS4A1",
                 "CD69"
                 )
                 foo[[1]] <- "CD14+ Macrophage";
                 foo[[2]] <- "LAMP3+ Mature DC";
                 foo[[3]] <- "FCER1A+ Dermal DC";
                 foo[[4]] <- "IL7R+ T4";
                 foo[[5]] <- "PRF1+ gdT";
                 foo[[6]] <- "IRF8+ DC";
                 foo[[7]] <- "CPA3+ MAST";
                 foo[[8]] <- "GZMH+ NKT";
                 foo[[9]] <- "GZMK+ T8";
                 foo[[10]] <- "FOXP3+ Treg";
                 foo[[11]] <- "MS4A1+ B";
                 foo[[12]] <- "CD69+ Activated T";

             } else if(model == "huBM"){
                 names(foo) <- c("MME","CD19","DNAJB9","ALAS2","PMP22","SMIM24","PDLIM1","CRHBP","IL3RA","FGR","GZMA","CD2","GNLY")
                 foo[[1]] <- "MME+ pre-B Cells";
                 foo[[2]] <- "CD19+ B Cells";
                 foo[[3]] <- "DNAJB9+ Plasma B Cells";
                 foo[[4]] <- "ALAS2+ Erythroblasts";
                 foo[[5]] <- "PMP22+ Erythroblasts";
                 foo[[6]] <- "SMIM24+ HSPC GMP";
                 foo[[7]] <- "SMIM24+ HSPC GMP";
                 foo[[8]] <- "CRHBP+ CD34 HSPC";
                 foo[[9]] <- "IL3RA+ Basophils";
                 foo[[10]] <- "FGR+ Monocytes";
                 foo[[10]] <- "GZMA+ T8 Cells";
                 foo[[10]] <- "CD2+ T Cells";
                 foo[[10]] <- "GNLY+ NK Cells";

             }

             if (scTransform){
                 subsetDF = seuratSeqGeq@assays[["SCT"]]@scale.data
             } else {
                 if(runBatchCorrection) {
                     subsetDF = seuratSeqGeq@assays[["integrated"]]@scale.data
                 }  else {
                     subsetDF = seuratSeqGeq@assays[["RNA"]]@scale.data
                 }
             }
             subsetList = seuratSeqGeq@active.ident
             subDF = rbind(subsetDF, subsetList)
             subDFT = as.data.frame(t(subDF))

             clusterTypesMatrix <- data.frame(matrix(ncol = length(ClassGeneSet_rd), nrow = max(as.list.data.frame(subsetList))))
             colnames(clusterTypesMatrix) <- ClassGeneSet_rd
             rownames(clusterTypesMatrix) <- paste(rownames(clusterTypesMatrix),"Cluster",sep="")
             orderedDF = subDFT[order(subDFT$subsetList),]
             for(gene in colnames(clusterTypesMatrix))
             {
                 # Test
                 # gene <- "IL7R"

                 print(gene)
                 lowerBound = -1
                 upperBound = -1
                 onLevel = 999
                 currentMedian = -10

                 for(row in 1:length(orderedDF$subsetList))
                 {
                     # row <- 442

                     currentLevel = orderedDF[row,"subsetList"]
                     if(onLevel== 999)
                     {
                         print(currentLevel)
                         onLevel = currentLevel
                         lowerBound = 0
                         upperBound = row
                     }
                     else if(currentLevel > onLevel){
                         upperBound = row - 1
                         differ = upperBound - lowerBound
                         currentMedian = sum(orderedDF[lowerBound:upperBound,gene])/differ
                         lowerBound = upperBound
                         index = currentLevel -1
                         clusterTypesMatrix[index, gene] = currentMedian
                         onLevel = currentLevel
                         print(currentLevel)
                     }
                 }
                 upperBound = row
                 currentMedian = median(orderedDF[lowerBound:upperBound,gene])
                 index = currentLevel
                 clusterTypesMatrix[index, gene] = currentMedian
                 onLevel = currentLevel
                 print(currentLevel)
             }

             clusterTypesMatrixDF = clusterTypesMatrix
             cellCalls = data.frame(matrix(nrow=length(rownames(clusterTypesMatrixDF)), ncol=1))
             colnames(cellCalls) = "CellCall"
             count = 1
             for(cluster in rownames(clusterTypesMatrixDF))
             {

                 #print(length(colnames(clusterTypesMatrixDF)))
                 if(length(colnames(clusterTypesMatrixDF)) < 1)
                 {
                     break
                 }

                 meen <- rowMeans(as.matrix(clusterTypesMatrixDF[cluster,]))
                 std <- abs(sd(as.matrix(clusterTypesMatrixDF[cluster,])))

                 tst <- meen + standard*std

                 if (max(clusterTypesMatrixDF[cluster,]) > tst) {

                     cellCalls[count,1] = foo[[names(which.max(clusterTypesMatrixDF[cluster,]))]]
                     #print(which.max(clusterTypesMatrixDF[cluster,]))
                     print(foo[[(which.max(clusterTypesMatrixDF[cluster,]))]])
                     #clusterTypesMatrixDF = clusterTypesMatrixDF[,-c(which.max(clusterTypesMatrixDF[cluster,]))]
                 } else {
                     cellCalls[count,1] = "Unknown"
                     print(cellCalls[count,1])
                 }
                 count = count + 1
             }
             currentCount=1
             for(cell in cellCalls$CellCall)
             {
                 rownames(clusterTypesMatrix)[[currentCount]] = paste0("Cluster ", currentCount, " ", cell)
                 currentCount = currentCount +1
             }

             lenDF = length(foo) -1
             df <- data.frame(matrix(unlist(foo), nrow=length(foo), byrow=T),stringsAsFactors=FALSE)
             dff <- data.frame(matrix(unlist(names(foo)), nrow=length(foo), byrow=T),stringsAsFactors=FALSE)
             dfff <- cbind(df, dff)
             colnames(dfff) <- c("Type","Gene")

             write.csv(rownames(clusterTypesMatrix), file = paste0(".CellType_Dict.csv"), quote = FALSE, row.names = F)

             XtraCol <- as.data.frame(rownames(clusterTypesMatrix))
             colnames(XtraCol) <- c("Type")
             tMatrix <- cbind(XtraCol, clusterTypesMatrix)
             hmDF <- gather(tMatrix, "Features", "Fill", 2:ncol(tMatrix)) # :nrow(clusterTypesMatrix))

             ## Create Figures ##
           if (runTSNE) {
               featPlot <- FeaturePlot(seuratSeqGeq, features = ClassGeneSet_rd)
               ggsave(file=paste0("CellTypes_", runID, ".png"), device= "png", plot= featPlot, height= 6, width= 9, units= "in", dpi=300)
             }
           cellTypeHM <- fig2(hmDF)
           ggsave(file=paste0("CellTypeHeatmap_", runID, ".png"), device= "png", plot= cellTypeHM, height=5, width=8, units= "in", dpi=300)
           print(" Saved Images ")

         } else {
             writeLines(paste0("Features Found: ",ClassGeneSet_rd))
             writeLines("We could not find enough marker genes to model your data's phenotypes.")
             writeLines("These are the markers we need to identify the phenotypes in your selected model:")
             writeLines(ClassGeneSet)
             clustNames <- paste0("Cluster_",sort(unique(seuratSeqGeq@meta.data$seurat_clusters)))
             write.csv(clustNames, file = paste0(".CellType_Dict.csv"), quote = FALSE, col.names = F, row.names = F)
         }

     } else {

            ## WE DON'T WANT CLASSIFICATION ##
            # ... but we still need a file to name clusters for SeqGeq... #
            subsetList = seuratSeqGeq@active.ident
            clusterTypesMatrix <- data.frame(matrix(ncol = 1, nrow = max(as.list.data.frame(subsetList))))
            colnames(clusterTypesMatrix) <- "ClustyClust"
            rownames(clusterTypesMatrix) <- paste("Cluster", rownames(clusterTypesMatrix), sep=" ")
            cellCalls = data.frame(matrix(nrow=length(rownames(clusterTypesMatrix)), ncol=1))
            colnames(cellCalls) = "CellCall"
            write.csv(rownames(clusterTypesMatrix), file = paste0(".CellType_Dict.csv"), quote = FALSE, col.names = F, row.names = F)
     }

  top10 <- seuratSeqGeq.markers %>% group_by(cluster) %>% top_n(n = 10, wt = avg_log2FC)

# option to print each UMAP for each assay this will incerase run time.
if(runWeightedNearestNeighbors) {
  adtlength <- length(adtParNames)
  apcaDim <- adtlength - 2

  # print("Running comparison Umap dimensionality reductions")
  # Removed these plots b/c of issues with plots sizes and saving of images. Possible solutions are to change output
  # size of the plots or wrap with TryCatch to generate error message file.
  # seuratSeqGeq <- RunUMAP(seuratSeqGeq, reduction = 'pca', dims = 1:30, assay = 'RNA',
  #                         reduction.name = 'rna.umap', reduction.key = 'rnaUMAP_')
  # seuratSeqGeq <- RunUMAP(seuratSeqGeq, reduction = 'apca', dims = 1:apcaDim, assay = 'ADT',
  #                         reduction.name = 'adt.umap', reduction.key = 'adtUMAP_')
  # p1 <- DimPlot(seuratSeqGeq, reduction = 'wnn.umap', label = TRUE,
  #               repel = TRUE, label.size = 1.5, label.box = F, pt.size=0.01) + NoLegend() + FontSize(x.text=1,y.text=1,x.title = 7, y.title = 1)
  # p2 <- DimPlot(seuratSeqGeq, reduction = 'rna.umap', label = TRUE,
  #               repel = TRUE, label.size = 1.5, label.box = F,pt.size=0.01) + NoLegend() + FontSize(x.text=1,y.text=1,x.title = 7, y.title = 1)
  # p3 <- DimPlot(seuratSeqGeq, reduction = 'adt.umap', label = TRUE,
  #               repel = TRUE, label.size = 1.5, label.box = F,pt.size=0.01) + NoLegend() + FontSize(x.text=1,y.text=1,x.title = 7, y.title = 1)
  #
  # umapComparison <- p1 + p2 + p3
  # # png(filename = paste0(outfileDirectory, "/Plot1_Umap_Comparisons_", runID, ".png"), height= 900, width= 1500, units= "px", bg = "transparent")
  # # cowplot::plot_grid(umapComparison)
  # # dev.off()
  # ggsave(filename= paste0("Plot1_Umap_Comparisons_", runID, ".png"), device= "png", plot= umapComparison, height= 900, width= 1500, units= "px", dpi= 300, bg = "transparent")

  if(logNormalizeBool){
    weightViolins <- VlnPlot(seuratSeqGeq, features = "RNA.weight", group.by = 'seurat_clusters', sort = TRUE, pt.size = 0.1) + NoLegend()
  }else{
    weightViolins <- VlnPlot(seuratSeqGeq, features = "SCT.weight", group.by = 'seurat_clusters', sort = TRUE, pt.size = 0.1) + NoLegend()
  }
  tryCatch({
    ggsave(filename = paste0("Plot2_Modality_Weights_ViolinPlot_", runID, ".png"), device= "png", plot= weightViolins, height= 900, width= 1500, units= "px", dpi= 300, bg = "transparent")
  }, error = function(e){
    message <- "Seurat was not able to save violins plots."
    ReturnErrorMessage(outfileDirectory, message)
  })


  # gene_plot = seuratSeqGeq.markers.RNA %>% filter(avg_log2FC > 1 ) %>%  group_by(cluster) %>% top_n(3) %$% gene %>% unique
  #
  # d1 <- seuratSeqGeq@meta.data
  # d2 <- as.data.frame(t(seuratSeqGeq@assays$ADT@data))
  # d3 <- as.data.frame(t(seuratSeqGeq@assays$RNA@data[gene_plot,]))
  # d <- cbind(d1,d2,d3)
  # adtPars <- colnames(d2)

  # adt heatmap
  # adt_plot = d %>%
  #   dplyr::group_by(seurat_clusters) %>%
  #   dplyr::summarize_at(.vars = c(adtPars, gene_plot), .funs = median) %>%
  #   tibble::remove_rownames() %>%
  #   tibble::column_to_rownames("seurat_clusters")
  # mRNA heatmap
  # mrna = t(adt_plot)[gene_plot, ]

  # p1 = ComplexHeatmap::Heatmap(t(adt_plot)[adtPars, ],
  #              col = inferno(100),
  #              name = "protein",
  #              use_raster = T,
  #              row_names_gp = grid::gpar(color = "black", fontsize = 3),
  #              column_names_rot = 0,
  #              show_heatmap_legend = TRUE,
  #              rect_gp = grid::gpar(col="grey",lwd=0.25),
  #              heatmap_legend_param = list(
  #                legend_direction="vertical",
  #                title_position="topcenter",
  #                title_gp=grid::gpar(fontsize = 5)
  #              )
  # )
  # p2 = ComplexHeatmap::Heatmap(t(scale(t(mrna))), name = "mRNA",use_raster = T,
  #              clustering_method_columns = 'average',
  #              col = inferno(100),
  #              column_names_gp = grid::gpar(color = "black", fontsize = 5),
  #              column_names_rot = 0,
  #              row_names_gp = grid::gpar(color = "black", fontsize = 3),
  #              rect_gp = grid::gpar(col="grey",lwd=0.25),
  #              show_heatmap_legend = TRUE,
  #              heatmap_legend_param = list(
  #                legend_direction="vertical",
  #                title_position = "topcenter",
  #                title_gp=grid::gpar(fontsize=5))
  # )

  # ht_list = p1 %v% p2
  # gb = grid.grabExpr(draw(ht_list, heatmap_legend_side="right"),wrap = TRUE)
  # pAll <- as_ggplot(gb)
  # ggsave(filename = paste0("ClusterHeatmap_", runID, ".png"), width=1500, height=900, unit="px", plot = pAll, device = "png", dpi=300, bg = "transparent")
  if(logNormalizeBool){
    RNAHeatmap <-  DoHeatmap(seuratSeqGeq, assay = "RNA",features = top10$gene, size=2) + theme(legend.key.size = unit(0.25,"cm"), text = element_text(size = 3))
    tryCatch({
      ggsave(filename = paste0("RNAHeatmap_", runID, ".png"), width=1500, height=900, unit="px", plot = RNAHeatmap, device = "png", dpi=300, bg = "transparent")
    }, error = function(e){
      message <- "Seurat was not able to save RNA Heatmap plots."
      ReturnErrorMessage(outfileDirectory, message)
    })
  }else{
    RNAHeatmap <-  DoHeatmap(seuratSeqGeq, assay = "SCT",features = top10$gene, size=2) + theme(legend.key.size = unit(0.25,"cm"), text = element_text(size = 3))
    tryCatch({
      ggsave(filename = paste0("RNAHeatmap_", runID, ".png"), width=1500, height=900, unit="px", plot = RNAHeatmap, device = "png", dpi=300, bg = "transparent")
    }, error = function(e){
      message <- "Seurat was not able to save RNA Heatmap plots."
      ReturnErrorMessage(outfileDirectory, message)
    })
  }
  ADTHeatmap <-  DoHeatmap(seuratSeqGeq, assay = "ADT", features = top10$gene, size=2) + theme(legend.key.size = unit(0.25,"cm"), text = element_text(size = 3))
  tryCatch({
    ggsave(filename = paste0("ADTHeatmap_", runID, ".png"), width=1500, height=900, unit="px", plot = ADTHeatmap, device = "png", dpi=300, bg = "transparent")
  }, error = function(e){
    message <- "Seurat was not able to save Protein Heatmap plots."
    ReturnErrorMessage(outfileDirectory, message)
  })
}

if(!runWeightedNearestNeighbors){
  if(runBatchCorrection){
    clusterHeatmap <-  DoHeatmap(seuratSeqGeq, assay = "integrated", features = top10$gene, size=2) + theme(legend.key.size = unit(0.25,"cm"), text = element_text(size = 3))
    tryCatch({
      ggsave(file=paste0("ClusterHeatmap_", runID, ".png"), device= "png", plot= clusterHeatmap, height= 900, width= 1500, units= "px", dpi=300, bg = "transparent")
    }, error = function(e){
      message <- "Seurat was not able to save Heatmap plots."
      ReturnErrorMessage(outfileDirectory, message)
    })
  }else{
    clusterHeatmap <-  DoHeatmap(seuratSeqGeq, features = top10$gene, size=2) + theme(legend.key.size = unit(0.25,"cm"), text = element_text(size = 3))
    tryCatch({
      ggsave(file=paste0("ClusterHeatmap_", runID, ".png"), device= "png", plot= clusterHeatmap, height= 900, width= 1500, units= "px", dpi=300, bg = "transparent")
    }, error = function(e){
      message <- "Seurat was not able to save Heatmap plots."
      ReturnErrorMessage(outfileDirectory, message)
    })
  }
}


#Finishing Up
writeLines(capture.output(sessionInfo()), paste0(outfileDirectory, "/RsessionInfo.txt"))
writeLines(" Finished: Script. Returning results back to SeqGeq.")
save(seuratSeqGeq, file = paste0("SeqGeq_Seurat_Object_", runID, ".rda"), compress = TRUE)
write.table(data.frame(1,2,3,5,8), file="fin.tsv", sep="  ", col.names=F, row.names = F)
#Clean up
rm(list=ls())
